theory Formula
imports
  "$AFP/Abstract-Rewriting/Abstract_Rewriting"
  Term
  Trs_Impl
  Option_Monad
begin

type_synonym ('f,'v,'t) exp = "('f, 'v \<times> 't) term"
type_synonym ('v,'t,'d) valuation = "'v \<times> 't \<Rightarrow> 'd"

type_synonym ('f,'v,'t) formula = "('f,'v,'t) exp list list"

abbreviation disj_atoms :: "('f,'v,'t) exp list \<Rightarrow> ('f,'v,'t) formula" where
  "disj_atoms \<equiv> (\<lambda> as. [as])"
abbreviation conj_atoms :: "('f,'v,'t) exp list \<Rightarrow> ('f,'v,'t) formula" where
  "conj_atoms \<equiv> (\<lambda> as. (map (\<lambda> a. [a]) as))"

abbreviation form_False :: "('f,'v,'t) formula" where "form_False \<equiv> [[]]"
abbreviation form_True  :: "('f,'v,'t) formula" where "form_True \<equiv> []"

definition form_and :: "('f,'v,'t) formula \<Rightarrow> ('f,'v,'t) formula \<Rightarrow> ('f,'v,'t) formula"
  (infix "\<and>\<^sub>f" 80) where [simp]: "\<phi> \<and>\<^sub>f \<psi> \<equiv> \<phi> @ \<psi>"

consts rename_vars :: "('a \<Rightarrow> 'b) \<Rightarrow> 'c \<Rightarrow> 'd"

definition rename_vars_exp :: "('v \<Rightarrow> 'w) \<Rightarrow> ('f,'v,'t) exp \<Rightarrow> ('f,'w,'t) exp" where
  "rename_vars_exp r = map_term id (map_prod r id)"

adhoc_overloading rename_vars rename_vars_exp

lemma rename_vars_Fun: "rename_vars r (Fun f ts) = Fun f (map (rename_vars r) ts)"
  unfolding rename_vars_exp_def by auto

fun rename_vars_formula :: "('v \<Rightarrow> 'w) \<Rightarrow> ('f,'v,'t) formula \<Rightarrow> ('f,'w,'t) formula" where
  "rename_vars_formula r cls = map (map (rename_vars r)) cls"

adhoc_overloading rename_vars rename_vars_formula

record ('f,'t,'d) "theory" =
  types :: "'f \<Rightarrow> 't list \<times> 't"
  type_values :: "'t \<Rightarrow> 'd set"
  bool :: "'t"
  to_bool :: "'d \<Rightarrow> bool"
  "interpret" :: "'f \<Rightarrow> 'd list \<Rightarrow> 'd"

locale "pretheory" =
  fixes T :: "('f,'t,'d) theory"
begin

  abbreviation types where "types \<equiv> theory.types T"
  abbreviation type_values where "type_values \<equiv> theory.type_values T"
  abbreviation bool where "bool \<equiv> theory.bool T"
  abbreviation to_bool where "to_bool \<equiv> theory.to_bool T"
  abbreviation "interpret" where "interpret \<equiv> theory.interpret T"

  definition "assignment \<alpha> \<equiv> \<forall>x t. \<alpha> (x,t) \<in> type_values t"

  fun type_of :: "('f,'v,'t) exp \<Rightarrow> 't option" where
    "type_of (Var v) = Some (snd v)" |
    "type_of (Fun f ts) = do {
        tys \<leftarrow> mapM type_of ts;
        let (tys',t) = types f;
        guard (tys = tys');
        Some t
      }"

  definition "exp e \<equiv> type_of e \<noteq> None"
  abbreviation "EXP \<equiv> Collect exp"

  definition "is_bool e \<equiv> type_of e = Some bool \<and> \<not> is_Var e"
  abbreviation "BOOL_EXP \<equiv> Collect is_bool"

  lemma is_bool_imp_exp[simp]: "is_bool e \<Longrightarrow> exp e" unfolding is_bool_def exp_def by auto

  lemma type_of_fun: assumes "type_of (Fun f ts) \<noteq> None"
    shows "type_of (Fun f ts) = Some (snd (types f))"
    and "mapM type_of ts = Some (fst (types f))"
  proof -
    obtain v where a:"type_of (Fun f ts) = Some v" using assms by auto
    have "(let (tys', t) = types f in
          guard (the (Option_Monad.mapM type_of ts) = tys') \<bind> (\<lambda>_. Some t)) =
          Some v" using a unfolding type_of.simps by auto
    hence "guard (the (Option_Monad.mapM type_of ts) = fst (types f))
          \<bind> (\<lambda>_. Some (snd (types f))) =
          Some v" by (simp add: prod.case_eq_if)
    thus "type_of (Fun f ts) = Some (snd (types f))"
     and "mapM type_of ts = Some (fst (types f))" using a by auto
  qed

  lemma type_of_rename_vars[simp]: "type_of (rename_vars r a) = type_of a"
  proof (induct a)
    case (Fun f ts)
    obtain tys' t where ty: "types f = (tys',t)" by force
    have id: "Option_Monad.mapM type_of (map (rename_vars r) ts) = Option_Monad.mapM type_of ts"
      using Fun by (induct ts, auto)
    show ?case using ty by (auto simp: rename_vars_Fun id)
  qed (simp add: rename_vars_exp_def)

  lemma is_bool_rename_vars[simp]: "is_bool (rename_vars r a) = is_bool a"
    unfolding is_bool_def type_of_rename_vars
    by(cases a,auto simp:rename_vars_exp_def)

  fun typelist_match where
    "typelist_match (Cons a as) (Cons b bs) = (a \<in> type_values b \<and> typelist_match as bs)" |
    "typelist_match [] [] = True" |
    "typelist_match _ _ = False"

  definition formula :: "('f,'v,'t) formula \<Rightarrow> bool" where
    "formula \<phi> \<equiv> \<forall>cl \<in> set \<phi>. \<forall>lit \<in> set cl. is_bool lit"

  lemma formula_rename_vars[simp]: "formula (rename_vars r f) = formula f" by (auto simp: formula_def)

  lemma formula_and[simp]: "formula (\<phi> \<and>\<^sub>f \<psi>) \<longleftrightarrow> formula \<phi> \<and> formula \<psi>" by (auto simp: formula_def)
  lemma formula_False[simp]: "formula form_False" by (auto simp: formula_def)
  lemma formula_True[simp]: "formula form_True" by (auto simp: formula_def)

  context
    fixes \<alpha> :: "('v,'t,'d) valuation"
  begin

  fun eval :: "('f,'v,'t) exp \<Rightarrow> 'd" where
    "eval (Var v) = \<alpha> v" |
    "eval (Fun f ts) = interpret f (map eval ts)"

  definition "eval_true = to_bool \<circ> eval"

  definition satisfies :: "('f,'v,'t) formula \<Rightarrow> bool" where [simp]:
    "satisfies \<phi> \<equiv> \<forall>cl \<in> set \<phi>. \<exists>lit \<in> set cl. eval_true lit"

  lemma form_True[simp]: "satisfies form_True" by simp

  lemma satisfies_andI[intro]: "satisfies \<phi> \<Longrightarrow> satisfies \<psi> \<Longrightarrow> satisfies (\<phi> \<and>\<^sub>f \<psi>)" by auto

  end

  context
    fixes \<alpha> :: "('v,'t,'d) valuation" and \<beta> :: "('w,'t,'d) valuation" and r :: "'v \<Rightarrow> 'w"
    assumes r: "\<And> x t. \<alpha> (x,t) = \<beta> (r x, t)"
  begin
  lemma eval_rename_vars: "eval \<beta> (rename_vars r a) = eval \<alpha> a"
  proof (induct a)
    case (Var x)
    show ?case by (cases x, auto simp: rename_vars_exp_def r)
  next
    case (Fun f ts)
    thus ?case by (auto simp: rename_vars_Fun intro: arg_cong[of _ _ "interpret f"])
  qed

  lemma eval_true_rename_vars[simp]: "eval_true \<beta> (rename_vars r \<phi>) = eval_true \<alpha> \<phi>"
    by (auto simp: eval_true_def eval_rename_vars)

  lemma satisfies_rename_vars[simp]: "satisfies \<beta> (rename_vars r \<phi>) = satisfies \<alpha> \<phi>" by auto
  end

  definition implies :: "('f,'v,'t) formula \<Rightarrow> ('f,'v,'t) formula \<Rightarrow> bool" (infix "\<longrightarrow>\<^sub>f" 60) where
    "\<phi> \<longrightarrow>\<^sub>f \<psi> \<equiv> \<forall> \<alpha>. assignment \<alpha> \<longrightarrow> formula \<phi> \<longrightarrow> formula \<psi> \<longrightarrow> satisfies \<alpha> \<phi> \<longrightarrow> satisfies \<alpha> \<psi>"

  lemma impliesD:
   "assignment \<alpha> \<Longrightarrow> formula \<phi> \<Longrightarrow> formula \<psi> \<Longrightarrow> \<phi> \<longrightarrow>\<^sub>f \<psi> \<Longrightarrow> satisfies \<alpha> \<phi> \<Longrightarrow> satisfies \<alpha> \<psi>"
    unfolding implies_def by auto

  lemma impliesI[intro]:
   "(\<And> \<alpha>. assignment \<alpha> \<Longrightarrow> formula \<phi> \<Longrightarrow> formula \<psi> \<Longrightarrow> satisfies \<alpha> \<phi> \<Longrightarrow> satisfies \<alpha> \<psi>)
     \<Longrightarrow> \<phi> \<longrightarrow>\<^sub>f \<psi>"
    unfolding implies_def by auto

end

locale "theory" = "pretheory" T for T :: "('f,'t,'d) theory" +
  assumes well_typed:
    "typelist_match ds (fst (types f)) \<Longrightarrow> interpret f ds \<in> type_values (snd (types f))"
begin

  lemma eval_types:
    assumes "type_of x = Some v"
    and "assignment \<alpha>"
    shows "eval \<alpha> x \<in> type_values v"
  using assms proof (induct x arbitrary: v)
    case (Fun f ls)
      obtain tps where tps: "fst (types f) = tps" by auto
      hence "mapM type_of ls = Some tps" using Fun.prems type_of_fun(2) by auto
      hence tlm: "typelist_match (map (eval \<alpha>) ls) (fst (types f))" unfolding tps
        using Fun.hyps assms(2)
      proof(induct "(map (eval \<alpha>) ls)" tps arbitrary: ls f rule: typelist_match.induct)
        case (1 a as b bs)
          obtain l ls' where ls:"ls = Cons l ls'" using "1.hyps"(2) by auto
          have typ_l_b : "type_of l = Some b" using "1.prems"(1) "1.hyps"(2) 
            unfolding ls "1.hyps"(2)[symmetric] by simp
          have as:"as = map (eval \<alpha>) ls'" using "1.hyps"(2) ls by auto
          have map:"Option_Monad.mapM type_of ls' = Some bs" using "1.prems" ls by auto
          have "a = (eval \<alpha> l)" using "1.hyps"(2) ls by auto
          hence "eval \<alpha> l \<in> type_values b" using "1.prems"(2) typ_l_b assms(2) ls by auto
          thus ?case unfolding list.map typelist_match.simps
            using "1.hyps"(1)[of ls',OF as map] "1.prems" list.set_intros(2) ls by auto
      qed(auto)
      show ?case using well_typed[OF tlm] unfolding "eval.simps"
        using Fun.prems(1) option.discI option.sel type_of_fun(1) by fastforce next
    case (Var var)
      have v:"v = snd var" using Var.prems(1) unfolding type_of.simps by auto
      show ?case using assms(2)
        unfolding eval.simps assignment_def v using prod.collapse by metis
  qed
end

type_synonym ('h,'f,'v,'t) theory_checker = 
  "'h \<Rightarrow> ('f,'v,'t) exp list \<Rightarrow> ('f,'v,'t) exp list \<Rightarrow> shows check"

locale pre_theory_checker = pretheory T
  for T :: "('f::show,'t::show,'d) theory"
  and theory_checker :: "('h::show,'f,'v::show,'t) theory_checker"
begin

definition formula_checker ::
    "'h list  \<Rightarrow> ('f, 'v \<times> 't) Term.term list list \<Rightarrow> ('f, 'v \<times> 't) Term.term list list \<Rightarrow> (char list \<Rightarrow> char list) + unit"
where
"formula_checker hints \<phi> \<psi> \<equiv>
  if [] \<in> set \<phi> then succeed else do {
     check_allm (\<lambda> cl. check (length cl = 1) (shows ''disjunctions in premises not (yet) supported'')) \<phi>;
     check (length hints = length \<psi>) (shows ''number of hints does not match number of conclusions'');
     let prems = map hd \<phi>;
     check_allm (\<lambda> (h,c). theory_checker h prems c <+? (\<lambda> s. shows ''The conclusion '' o shows c o shows '' does not follow from the premises '' o shows prems o shows '' via the hints '' o shows h o shows ''\<newline>'' o s)) (zip hints \<psi>)
   }"
end

locale theory_checker =
  "theory" T + pre_theory_checker T theory_checker
  for T :: "('f::show,'t::show,'d) theory"
  and theory_checker :: "('h::show,'f,'v::show,'t) theory_checker" +
  assumes theory_checker: "isOK(theory_checker h prems concls)
    \<Longrightarrow> set prems \<subseteq> BOOL_EXP
    \<Longrightarrow> set concls \<subseteq> BOOL_EXP
    \<Longrightarrow> assignment \<alpha>
    \<Longrightarrow> \<forall>prem \<in> set prems. eval_true \<alpha> prem
    \<Longrightarrow> \<exists>concl \<in> set concls. eval_true \<alpha> concl"
begin

lemma theory_checker_implies: assumes "isOK(theory_checker h prems concls)"
   "set prems \<subseteq> BOOL_EXP"
   "set concls \<subseteq> BOOL_EXP"
   shows "conj_atoms prems \<longrightarrow>\<^sub>f disj_atoms concls"
   using theory_checker[OF assms] by auto

lemma formula_checker: assumes ok: "isOK(formula_checker hints \<phi> \<psi>)"
  shows "\<phi> \<longrightarrow>\<^sub>f \<psi>"
proof
  fix \<alpha>
  assume \<alpha>: "assignment \<alpha>"
     and \<phi>: "formula \<phi>"
     and \<psi>: "formula \<psi>"
     and \<alpha>\<phi>: "satisfies \<alpha> \<phi>"
  show "satisfies \<alpha> \<psi>" unfolding satisfies_def
  proof
    fix cl
    assume cl: "cl \<in> set \<psi>"
    then obtain i where i: "i < length \<psi>" and cli: "cl = \<psi> ! i"
      unfolding set_conv_nth by auto
    from \<alpha>\<phi> have empty: "[] \<notin> set \<phi>" by auto
    note ok = this ok[unfolded formula_checker_def]
    from ok have "length hints = length \<psi>" by auto
    with i have "(hints ! i, \<psi> ! i) \<in> set (zip hints \<psi>)" unfolding set_zip
      by auto
    then obtain h where "(h, cl) \<in> set (zip hints \<psi>)" unfolding cli by blast
    with ok have tc': "isOK (theory_checker h (map hd \<phi>) cl)" by auto
    show "Bex (set cl) (eval_true \<alpha>)"
    proof (rule theory_checker[OF tc' _ _ \<alpha>])
      show "set cl \<subseteq> BOOL_EXP" using cl \<psi> by (auto simp: formula_def)
      show "set (map hd \<phi>) \<subseteq> BOOL_EXP" using empty \<phi> by (induct \<phi>, auto simp: formula_def)
      from ok have "\<And> x. x \<in> set \<phi> \<Longrightarrow> length x = Suc 0" by auto
      thus "\<forall>prem \<in> set (map hd \<phi>). eval_true \<alpha> prem" using \<alpha>\<phi> unfolding set_map
      proof (induct \<phi>)
        case (Cons c cls)
        from Cons(2)[of c] obtain d where c: "c = [d]" by (cases c, cases "tl c", auto)
        from Cons show ?case unfolding c by auto
      qed auto
    qed
  qed
qed

end

end