theory Safety_Proof_Parser
imports
  Safety_Checker
  "$AFP/XML/Xmlt"
begin

datatype ('f,'v,'t,'l,'tr,'h) safety_cert_problem =
   Safety_Cert_Problem "('f,'v,'t,'l,'tr) lts_impl \<times> 'l" "('f,'v,'t,'l,'tr,'h)art_impl"


type_synonym ('f,'v,'t) exp_parser = "'v xmlt \<Rightarrow> ('f,'v,'t) exp xmlt"

definition formula_parser :: "('f,'a,'t) exp xmlt \<Rightarrow> ('f,'a,'t) formula xmlt" where
  "formula_parser ep = Xmlt.singleton ''formula'' (Xmlt.many ''conjunction'' ep id) conj_atoms"

fun drop_xml_attributes :: "xml \<Rightarrow> xml" where
  "drop_xml_attributes (XML n atts cs) = (XML n [] cs)"
| "drop_xml_attributes x = x"

context
  fixes node :: "'l xmlt"
  and trans_id :: "'tr xmlt"
  and exp :: "('f,'v,'t) exp xmlt"
  and texp :: "('f,'v trans_var,'t) exp xmlt"
  and hint :: "'h xmlt"
begin

abbreviation invariant where "invariant \<equiv> formula_parser exp"
abbreviation "transition_formula \<equiv> formula_parser texp"

definition transition :: "('tr \<times> ('f,'v,'t,'l) transition) xmlt" where
  "transition = Xmlt.tuple4 ''transition''
    trans_id
    (Xmlt.singleton ''source'' node id)
    (Xmlt.singleton ''target'' node id)
    transition_formula
    (\<lambda> i s t f. (i, Transition s f t))
    "

definition lts :: "('f,'v,'t,'l,'tr) lts_impl xmlt" where
  "lts = Xmlt.many1 ''program'' 
     (Xmlt.singleton ''initial'' node id)
     transition
     Lts_Impl"

definition safety_program :: "(('f,'v,'t,'l,'tr) lts_impl \<times> 'l) xmlt" where
  "safety_program = Xmlt.pair ''safetyProgram''
     lts
     (Xmlt.singleton ''error'' node id)
     Pair"

definition art_node_id :: "art_node xmlt" where
  "art_node_id = Xmlt.text ''artNodeId''"

definition hints :: "'h list xmlt" where
  "hints = Xmlt.many ''hints'' hint id"


definition art_node :: "('f,'v,'t,'l,'tr,'h list) art_node_impl xmlt" where
  "art_node = Xmlt.tuple4 ''artNode''
     art_node_id
     invariant
     (Xmlt.singleton ''node'' node id)
     (Xmlt.options [
       (''children'', Xmlt.many ''children'' 
         (Xmlt.triple ''child'' trans_id art_node_id hints (\<lambda> t n h. (t,n,h))) Children_Edge),
       (''coverEdge'', Xmlt.pair ''coverEdge'' art_node_id hints Cover_Edge)])
     Art_Node"

definition art_safety_proof :: "('f, 'v, 't, 'l, 'tr, 'h list) art_impl xmlt" where
  "art_safety_proof = Xmlt.triple ''impact''
     art_node_id
     (Xmlt.many ''artNodes'' art_node id)
     (Xmlt.many ''errorHints'' (Xmlt.pair ''errorHint'' art_node_id hints Pair) id)
     (\<lambda> init ns hs. \<lparr>art_impl.initial_node = init, nodes = ns, error_hints = hs\<rparr>)"

definition safety_proof :: "('f,'v,'t,'l,'tr,'h list) art_impl xmlt" where
  "safety_proof = Xmlt.singleton ''safetyProof'' art_safety_proof id"

definition safety_certification_problem :: "('f,'v,'t,'l,'tr,'h list) safety_cert_problem xmlt" where
  "safety_certification_problem =
    Xmlt.pair ''safetyProblem'' safety_program safety_proof Safety_Cert_Problem"

primrec
  xmldoc_safety_cert_problem ::
    "xmldoc \<Rightarrow> string +\<^sub>\<bottom> ('f, 'v, 't, 'l, 'tr, 'h list) safety_cert_problem"
where
  "xmldoc_safety_cert_problem (XMLDOC header xml) =
   safety_certification_problem (drop_xml_attributes xml)"


definition parse_safety_cert_problem ::
  "string \<Rightarrow> string +\<^sub>\<bottom> ('f, 'v, 't, 'l, 'tr, 'h list) safety_cert_problem"
where
  "parse_safety_cert_problem s \<equiv>
    case doc_of_string s of
      Inl x \<Rightarrow> error x
    | Inr y \<Rightarrow> xmldoc_safety_cert_problem y"

end

type_synonym node = string
type_synonym trans_id = string
type_synonym variable = string

definition node :: "node xmlt" where
  "node = Xmlt.text ''nodeIdentifier''"

definition trans_id :: "trans_id xmlt" where "trans_id = Xmlt.text ''transitionId''"

definition variable_name :: "variable xmlt" where
  "variable_name = Xmlt.singleton ''variableName''
     (Xmlt.text ''variableIdentifier'') id"

definition variable :: "variable xmlt" where
  "variable = Xmlt.singleton ''variable'' variable_name id"

definition trans_variable :: "variable trans_var xmlt" where
  "trans_variable = Xmlt.singleton ''variable''
    (Xmlt.options [
      (''variableName'', Xmlt.change variable_name Pre),
      (''post'', Xmlt.singleton ''post'' variable_name Post)
    ]) id"

end
